function dbg(strDebugText)
	if (g_debugprint) then print(strDebugText) end
	if (g_debuglog) then C4:ErrorLog(strDebugText) end
end
  
--------------------------
-- Global Variable Init --
--------------------------

tEmitCommands = {}
g_debugprint = true
g_debuglog = false
g_SerialBuffer = ''

g_SlowTimer = nil
g_VolUpTimer = nil
g_VolDownTimer = nil

g_currentInput = 1
g_volumeString = ''
g_soundMode = ''
g_powerStatus = ''
g_currentOutput = ''

g_asyncPort = nil

intraCommandDelayValue = Properties['Command Delay-Milliseconds']
intraSetInputCommandDelayValue = Properties['Set Input Command Delay-Milliseconds']
g_XMLString = ''
g_roomID = ''

g_inputVolume = {[1] = 0, [2] = 0, [3] = 0, [4] = 0, [5] = 0, [6] = 0, [7] = 0, [8] = 0,
			  [9] = 0, [10] = 0, [11] = 0, [12] = 0, [13] = 0, [14] = 0, [15] = 0, [16] = 0,
			  [17] = 0, [18] = 0, [19] = 0, [20] = 0, [21] = 0}

g_inputMute = {[1] = false, [2] = false, [3] = false, [4] = false, [5] = false, [6] = false, [7] = false, [8] = false,
			  [9] = false, [10] = false, [11] = false, [12] = false, [13] = false, [14] = false, [15] = false, [16] = false,
			  [17] = false, [18] = false, [19] = false, [20] = false, [21] = false}

			  
-- Required to exist
CMDS = {}

MAPPINGS = {
	['POWER_TOGGLE'] = 'PWRTGL',
	['ON'] = 'PWRONN',
	['OFF'] = 'PWROFF',
	['RESTART'] = 'SYSRST',
	['SET_MUTE_ON'] = 'MUTONN',
	['SET_MUTE_OFF'] = 'MUTOFF',
	['TOGGLE_MUTE'] = 'MUTTGL',
	['SET_VOL_UP'] = 'VOLUPP',
	['SET_VOL_DOWN'] = 'VOLDWN',
	['PULSE_VOL_UP'] = 'VOLUPP',
	['PULSE_VOL_DOWN'] = 'VOLDWN',
	['VOLUME_DIRECT'] = 'VOL0',
	['INPUT'] = 'INPP',
	['SOUND_MODE'] = 'SDM',
	['BASS_LEVEL_DIRECT'] = 'BAS',
	['TREBLE_LEVEL_DIRECT'] = 'TRB',
	['VERBOSE_RESPONSE_ENABLED'] = 'VERBON',
	['VERBOSE_RESPONSE_DISABLED'] = 'VERBOF',
	['VERBOSE_RESPONSE_MAX'] = 'VERBN2',
	['FRONT_PANEL_BRIGHTNESS_AUTO'] = 'FPBAUT',
	['FRONT_PANEL_BRIGHTNESS_LOW'] = 'FPBLOW',
	['FRONT_PANEL_BRIGHTNESS_MEDIUM'] = 'FPBMED',
	['FRONT_PANEL_BRIGHTNESS_HIGH'] = 'FPBHGH',
	['SET_LIPSYNC_DELAY'] = 'LPSD',
	['DOLBY_DRC_OFF'] = 'DLBDRF',
	['DOLBY_DRC_ON'] = 'DLBDRN',
	['DOLBY_SCS_OFF'] = 'DLBCSF',
	['DOLBY_SCS_ON'] = 'DLBCSN',
	['DTS_DRC_OFF'] = 'DTSDRF',
	['DTS_DRC_ON'] = 'DTSDRN',
	['DTS_DIALOG_CONTROL'] = 'DTSDG',
	['ASIP'] = 'ASIP',
	['DTSDG'] = 'DTSDG',
	['DLBDRN'] = 'DLBDRN',
	['DLBCSF'] = 'DLBCSF',
	['DLBCSN'] = 'DLBCSN',
	['DTSDRF'] = 'DTSDRF',
	['DTSDRN'] = 'DTSDRN',
	['STSPWR'] = 'STSPWR',
	['STSMUT'] = 'STSMUT',
	['STSINP'] = 'STSINP',
	['STSVOL'] = 'STSVOL',
	['STSMOD'] = 'STSMOD',
	['STSFPB'] = 'STSFPB',
	['STSLPS'] = 'STSLPS',
	['STSDLD'] = 'STSDLD',
	['STSDLC'] = 'STSDLC',
}

-- Table of commands with delay timer
tSlowCommands = {
	['PWR'] = Properties['Power Delay-Milliseconds'],
	['SYS'] = Properties['Restart Delay-Milliseconds'],
	['INP'] = Properties['Set Input Command Delay-Milliseconds'],
	['SDM'] = Properties['Set Input Command Delay-Milliseconds'],
}

-- Maps ASCII Command to IR Code ID
tIRCodes = {
	['PWRTGL'] = 21000,
	['PWRONN'] = 21001,
	['PWROFF'] = 21002,
	['MUTTGL'] = 21003,
	['MUTONN'] = 21004,
	['MUTOFF'] = 21005,
	['VOLUPP'] = 21006,
	['VOLDWN'] = 21007,
	['INPP01'] = 21008,
	['INPP02'] = 21009,
	['INPP03'] = 21010,
	['INPP04'] = 21011,
	['INPP05'] = 21012,
	['INPP06'] = 21013,
	['INPP07'] = 21014,
	['INPP08'] = 21015,
	['INPP09'] = 21016,
	['INPP10'] = 21017,
	['INPP11'] = 21018,
	['INPP12'] = 21019,
	['INPP13'] = 21020,
	['INPP14'] = 21021,
	['INPP15'] = 21022,
	['INPP16'] = 21023,
	['INPP17'] = 21024,
	['INPP18'] = 21025,
	['INPP19'] = 21026,
	['INPP20'] = 21027,
	['INPP21'] = 21028,
	['ANATGL'] = 21029,
	['DGITGL'] = 21030,
	['SDMAUT'] = 21031,
	['SDMDIR'] = 21032,
	['SDMSTR'] = 21033,
	['SDMANA'] = 21034,
	['SDMPAR'] = 21035,
	['SDMNGT'] = 21036,
	['SDMDTS'] = 21037,
	['SDMDLB'] = 21038,
	['SDM2D1'] = 21039,
	['SDM3D1'] = 21040,
	['SDM5D1'] = 21041,
	['SDM7D1'] = 21042,
	['FPBAUT'] = 21043,
	['FPBLOW'] = 21044,
	['FPBMED'] = 21045,
	['FPBHGH'] = 21046,
	['BASSM'] =  21047,
	['BASSP'] =  21048,
	['TRBM'] =   21049,
	['TRBP'] =   21050,
	['DLBDRF'] = 21051,
	['DLBDRN'] = 21052,
	['DLBCSF'] = 21053,
	['DLBCSN'] = 21054,
	['DTSDRF'] = 21055,
	['DTSDRN'] = 21056,
	['DTSDCM'] = 21057,
	['DTSDCP'] = 21058,
}

-- Maps Surround ID to ASCII sound mode
tSurroundCommands = {
[1]="AUT",
[2]="DIR",
[3]="STR",
[4]="ANA",
[5]="DLB",
[6]="DTS",
[7]="2D1",
[8]="3D1",
[9]="5D1",
[10]="7D1",
[11]="PAR",
[12]="NGT",
[13]="DOLBY_DRC_OFF",
[14]="DOLBY_DRC_ON",
[15]="DTS_DRC_OFF",
[16]="DTS_DRC_ON",
[17]="DOLBY_SCS_OFF",
[18]="DOLBY_SCS_ON",
}

--------------------
-- On Event Calls --
--------------------

-- Called whenever a property is changed
function OnPropertyChanged(strProperty)
  if (strProperty == 'Debug Mode') then
    if (Properties[strProperty] == 'Off') then
      g_debugprint = false
      g_debuglog = false
      g_DebugTimer = C4:KillTimer(g_DebugTimer)
      g_DebugTimer = 0
    end
    if (Properties[strProperty] == 'Print') then
      g_debugprint = true
      g_debuglog = false
      startDebugTimer()
    end
    if (Properties[strProperty] == 'Log') then
      g_debugprint = false
      g_debuglog = true
      startDebugTimer()
    end
    if (Properties[strProperty] == 'Print and Log') then
      g_debugprint = true
      g_debuglog = true
      startDebugTimer()
    end

  elseif (strProperty == 'Command Delay-Milliseconds') then
    intraCommandDelayValue = tonumber(Properties['Command Delay-Milliseconds'])
  elseif (strProperty == 'Power Delay-Milliseconds') then
    powerOnDelayValue = tonumber(Properties['Power Delay-Milliseconds'])
  elseif (strProperty == 'Set Input Command Delay-Milliseconds') then
    intraSetInputCommandDelayValue = tonumber(Properties['Set Input Command Delay-Milliseconds'])
  elseif (strProperty == 'Control Mode') then
    dbg('Control Mode Changed to: ' ..Properties["Control Mode"])

  elseif (strProperty == 'Send Command') then
	dbg('Sending Command: ' ..string.upper(Properties[strProperty]))
    table.insert(tEmitCommands, string.upper(Properties[strProperty]))
    idEmitCommandsTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS', false)
  elseif (strProperty == 'Director IP' and Properties['Control Mode'] == 'TCP') then
	CreateNewTCPServer()
  end
end

-- Called whenever a timer expires
function OnTimerExpired(idTimer)
	-- COMMAND TIMER
	if (idTimer == idEmitCommandsTimer) then
		if (#tEmitCommands > 0) then
			for k, cmd in pairs(tEmitCommands) do
				local cmdPrefix = string.sub(cmd, 1, 3)
				local cmdIsSlow = table.contains(tSlowCommands, cmdPrefix)
				if not(cmdIsSlow and g_SlowTimer ~= nil) then
				    dbg('Now Executing On Device: '..cmd)
				    
					-- Using the Control4 DriverWorks API to send commands to the device
					-- https://snap-one.github.io/docs-driverworks-api/#serial-and-network-interface-sendtonetwork

				    local controlMode = Properties['Control Mode']
				    if (controlMode == 'Serial') then

						-- SendToSerial Documentation
						-- Simple function which sends the command out serial port on binding 1 and adds the \r terminator to the end of the command being sent. This API should not be invoked during OnDriverInit.
						
						-- Available from 1.6.0
						-- Signature
						-- C4:SendToSerial(idBinding, strData)
						
						-- Parameter	Description
						-- num	Binding ID of the network interface to send on
						-- str	Data to send out the specified network interface
						-- Returns
						-- None
						
						-- Usage Note:
						-- Serial data to be sent can contain NULL characters. NULL (‘\0’) is a valid character and Lua strings handle embedded NULL characters

					   C4:SendToSerial(1, cmd)
					   dbg('Command sent via Serial')

				    --elseif (controlMode == 'UDP') then
				    --   local cmdHex = '06 F9 19 FF' .. StringToHex(cmd) .. '00 00'
				    --   C4:SendToNetwork(6002, Properties['UDP Port'], tohex(cmdHex))
				    --   dbg('Command sent via UDP Network')
				    elseif (controlMode == 'TCP') then
					   local cmdHex = StringToHex(cmd)

						-- SendToNetwork Documentation
						-- Function which sends an HTTP request to a network binding / port. This API should not be invoked during OnDriverInit.

						-- Available from 1.6.0
						-- Signature
						-- C4:SendToNetwork(idBinding, nPort, strData)

						-- Parameter	Description
						-- num	Binding ID of the network interface to send on
						-- num	Network port to send on
						-- str	Data to send out the specified network interface
						-- Returns
						-- None

						-- Usage Note:
						-- Data to be sent can contain NULL characters. NULL (‘\0’) is a valid character and Lua strings handle embedded NULL characters.

						-- Example
						-- C4:SendToNetwork(6001, 80, g_URLPacket)

						finalCommand = tohex(cmdHex)
						C4:SendToNetwork(6003, Properties['Control Port'], finalCommand)
						dbg(finalCommand .. ' sent via TCP Network')

				    elseif (controlMode == 'IR') then

						-- SendIR Documentation
						-- Function called from DriverWorks driver to send an IR Code. This API should not be invoked during OnDriverInit.
						
						-- Available from 1.6.0
						-- Signature
						-- C4:SendIR(idBinding,idIRCode)
						
						-- Parameter	Description
						-- num	idBinding: IR Binding ID to send the IR Code.
						-- num	idIRCode: ID of the IR Code to send from .c4i
						-- num	idBinding:Proxy Binding ID. (optional)
						-- Returns
						-- None
						
						-- Example (There isn't an actual usage documented)
						-- C4:SendIR(1, 22)

						-- Usage Note
						-- The IR code to send must be declared as an <ircode\> in the <irsection\> section of the driver file.

						if (tIRCodes[cmd] ~= nil) then
								C4:SendIR(2, tonumber(tIRCodes[cmd]))
								dbg('IR Code ID Match: (' ..tIRCodes[cmd] ..')')
								dbg('IR Code ID Match: ' ..tIRCodes[cmd])
						else
							dbg("Could Not Execute Command: " ..cmd.. " - No Matching IR Code ID")
						end
				    end
				    table.remove(tEmitCommands, k)

				    if (cmdIsSlow) then
					   g_SlowTimer = C4:AddTimer(tonumber(tSlowCommands[cmdPrefix]), 'MILLISECONDS')  -- Create new timer to repeat
				    end
				    idEmitCommandsTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS')  -- Create new timer to repeat
				    return
				end
				idEmitCommandsTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS') -- Create new timer to repeat
			end
	   else
		  idEmitCommandsTimer = nil  -- Get rid of the timer
	   end

	-- POWER TIMER
	elseif (idTimer == g_SlowTimer) then
		g_SlowTimer = nil

	-- VOLUME HOLD TIMER
	elseif (idTimer == g_VolUpTimer) then
		ExecuteCommand('SET_VOL_UP')
		g_inputVolume[g_currentInput] = math.Clamp(tonumber(g_inputVolume[g_currentInput]) + 1, 1, 99)
		g_VolUpTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS')
		C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
		C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
		C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
	elseif (idTimer == g_VolDownTimer) then
		ExecuteCommand('SET_VOL_DOWN')
		g_inputVolume[g_currentInput] = math.Clamp(tonumber(g_inputVolume[g_currentInput]) - 1, 1, 99)
		g_VolDownTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS')
		C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
		C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
		C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
	-- DEBUG TIMER
	elseif (idTimer == g_DebugTimer) then
		dbg('Turning Debug Mode back to Off [default] (timer expired)')
		C4:UpdateProperty('Debug Mode', 'Off')
		g_debugprint = false
		g_debuglog = false
		g_DebugTimer = C4:KillTimer(g_DebugTimer)
		return
	end
end

-- Called whenever a connection status has changed
function OnConnectionStatusChanged(idBinding, nPort, strStatus)
	dbg('Connection Status Changed: ' ..tostring(idBinding).. ' ' ..tostring(nPort).. ' ' ..tostring(strStatus))
	if (idBinding == 6003) then
		if (strStatus == 'OFFLINE') then
			C4:UpdateProperty('Network Ready','false')
		elseif (strStatus == 'ONLINE') then
			C4:UpdateProperty('Network Ready','true')
		end
	end
end

-- Called whenever a control binding has changed
function OnBindingChanged(idBinding, class, bIsBound)
	dbg('Binding Changed: ' ..tostring(idBinding).. ' ' ..tostring(class).. ' ' ..tostring(bIsBound))
	if (class == 'RS_232') then
		C4:UpdateProperty('Serial Ready', tostring(bIsBound))
	end
	if (class == 'IR_OUT') then
		C4:UpdateProperty('IR Ready', tostring(bIsBound))
	end
end

-- Called On Driver Destroyed
function OnDriverDestroyed()
	if (g_DebugTimer) then
		g_DebugTimer = C4:KillTimer(g_DebugTimer)
	end
	if (idEmitCommandsTimer) then
		idEmitCommandsTimer = C4:KillTimer(idEmitCommandsTimer)
		idEmitCommandsTimer = nil
	end
end

-------------------
-- Received From --
-------------------

function ReceivedFromProxy(idBinding, strCommand, tParams)
    dbg("ReceivedFromProxy: " ..strCommand)
    dbg("idBinding: " ..idBinding)

	if (tParams == nil) then tParams = {} end

	if (type(CMDS[strCommand]) == 'function') then
		CMDS[strCommand](strCommand, tParams)
		dbg('Executing Defined CMD Function: ' ..strCommand)
	end
end

function ReceivedFromNetwork(idBinding, nPort, strData)
  dbg('Received something from network on binding ' ..idBinding)
  g_SerialBuffer = g_SerialBuffer .. strData
  C4:UpdateProperty('Network Ready', 'true')

  local buflen, pos, start, len, data, lastpos
  buflen = string.len(g_SerialBuffer)
  pos = 1
  lastpos = 0 -- Tracks last 'used' byte in g_SerialBuffer, so we will not cut partial messages...

  while (pos < buflen) do
    -- If completed message received, process it.
      len = string.find(g_SerialBuffer,'\r',pos)
      if (len==nil) then break end

    -- Verify we have at least *len* to pull from g_SerialBuffer from pos to end...
    if (len <= buflen) then
      data = string.sub(g_SerialBuffer, pos, len - 1)
      pos = len+1
      lastpos = pos
      ProcessMessage(data)
    else
      -- Leaving with partial buffer...
      g_SerialBuffer = string.sub(g_SerialBuffer, lastpos)
      return
    end
  end
  g_SerialBuffer = string.sub(g_SerialBuffer, lastpos)
end

function ReceivedFromSerial(idBinding, strData)
    dbg('Received something from serial on binding ' .. idBinding)
    g_SerialBuffer = g_SerialBuffer .. strData
    C4:UpdateProperty('Serial Ready', 'true')
	
    local buflen, pos, start, len, data, lastpos
    buflen = string.len(g_SerialBuffer)
    pos = 1
    lastpos = 0 -- Tracks last 'used' byte in g_SerialBuffer, so we will not cut partial messages...

    while (pos < buflen) do
    -- If completed message received, process it.
      len = string.find(g_SerialBuffer,'\r',pos)
      if (len==nil) then break end

    -- Verify we have at least *len* to pull from g_SerialBuffer from pos to end...
      if (len <= buflen) then
        data = string.sub(g_SerialBuffer, pos, len - 1)
        pos = len+1
        lastpos = pos
        ProcessMessage(data)
      else
      -- Leaving with partial buffer...
        g_SerialBuffer = string.sub(g_SerialBuffer, lastpos)
        return
      end
    end
    g_SerialBuffer = string.sub(g_SerialBuffer, lastpos)
  end

----------------------
-- Process Commands --
----------------------

function ExecuteCommand(strCommandCode, tParams)
	-- Get the Mapping Code and Send to Device Emit Queue
	print(strCommandCode, tParams)
	local deviceCode = MAPPINGS[strCommandCode]
	if deviceCode == nil then print("MAPPING: '"..strCommandCode.."' NOT FOUND") return end

	-- Prevent commands from stacking (e.g. on, off, on, off)
	for k, v in pairs(tEmitCommands) do
		if (v == deviceCode) then return end
	end

	-- Add Parameters to Device Code
	if (tParams ~= nil and tParams ~= {}) then
		deviceCode = deviceCode .. tParams
		print(deviceCode)
	end

	-- Add Command to Emit Queue
	table.insert(tEmitCommands, deviceCode)
	if idEmitCommandsTimer == nil then
	  idEmitCommandsTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS', false)
	end

	dbg("Emit: " .. deviceCode)
end

function ProcessMessage(retMessage)
    -- Processes incomming Serial and Network messages
    message = retMessage
    message = string.gsub(message, ' ', '')
    message = string.gsub(message, '\r', '')
    message = string.gsub(message, '\n', '')
    dbg('MESSAGE: .' ..message..'.')
    
    if (string.find(message, 'OK') ~= nil) then
	   local tResponses = {'POWER', 'MUTE', 'SM', 'VOL0', 'INPUT'}
	   for _, option in pairs(tResponses) do
		  local starti, endi = string.find(message, option)
		  if (starti ~= nil and endi ~= nil) then 
			 
			 local response = string.sub(message, starti, #message)
			 local value = string.sub(message, endi +1, #message)
			 dbg('OPTION: .' ..option..'.')
			 dbg('RESPONSE: .' ..response..'.')
			 dbg('VALUE: .' ..value..'.')
			 
			if (response == 'POWERON') then
				g_powerStatus = 'ON'
				dbg('* POWERON Condition *')
				C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 7001}, 'NOTIFY')
				C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 4000}, 'NOTIFY')
				C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 2000}, 'NOTIFY')
			elseif (response == 'POWEROFF') then
				g_powerStatus = 'OFF'
				dbg('* POWEROFF Condition *')
				C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 7001}, 'NOTIFY')
				C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 4000}, 'NOTIFY')
				C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 2000}, 'NOTIFY')
			elseif (response == 'MUTEON') then
				g_inputMute[g_currentInput] = true
				dbg('* MUTEON Condition *')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
			elseif (response == 'MUTEOFF') then
				g_inputMute[g_currentInput] = false
				dbg('* MUTEOFF Condition *')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
			elseif (response == 'MUTETGL') then
				g_inputMute[g_currentInput] = not g_inputMute[g_currentInput]
				dbg('* MUTETGL Condition *')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
				C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
			elseif (option == 'INPUT') then
				dbg('* INPUT Condition *')
				g_currentInput = string.sub(value, 2, #value)
				C4:SendToProxy(5001, 'SET_INPUT_OUTPUT', {INPUT = g_currentInput, OUTPUT = 7001}, 'NOTIFY')				
				C4:SendToProxy(5001, 'SET_INPUT_OUTPUT', {INPUT = g_currentInput, OUTPUT = 4000}, 'NOTIFY')				
				C4:SendToProxy(5001, 'SET_INPUT_OUTPUT', {INPUT = g_currentInput, OUTPUT = 2000}, 'NOTIFY')				
			elseif (option == 'SM') then
				g_soundMode = value
				C4:SendToProxy(5001, 'SET_INPUT_OUTPUT', {SOUND_MODE = g_soundMode, OUTPUT = 7001}, 'NOTIFY')
				C4:SendToProxy(5001, 'SET_INPUT_OUTPUT', {SOUND_MODE = g_soundMode, OUTPUT = 4000}, 'NOTIFY')
				C4:SendToProxy(5001, 'SET_INPUT_OUTPUT', {SOUND_MODE = g_soundMode, OUTPUT = 2000}, 'NOTIFY')
			elseif (option == 'VOL0') then
				dbg('* VOL0 Condition *')
				g_inputVolume[g_currentInput] = math.Clamp(tonumber(value), 1, 99)
				C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
				C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
				C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
			end
			UpdateStatusProperties()
		  end
	   end
    end

    local strCommand = string.sub(message, 1, 3)
    local strParam = string.sub(message, 4, 6)
end

function startDebugTimer()
  if (g_DebugTimer) then
    g_DebugTimer = C4:KillTimer(g_DebugTimer)
  end
  g_DebugTimer = C4:AddTimer(90, 'MINUTES')
end

-----------------------
-- COMMAND FUNCTIONS --
-----------------------

function CMDS.LUA_ACTION(tParams,strCommand)
	if tParams ~= nil then
		for cmd,cmdv in pairs(tParams) do
			if cmd == 'ACTION' then
				ExecuteCommand(cmdv)
			end
		end
	end
end

function CMDS.ON(strCommand, tParams)
	g_powerStatus = 'ON'
	ExecuteCommand(g_powerStatus)
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.OFF(strCommand, tParams)
	g_powerStatus = 'OFF'
	ExecuteCommand(g_powerStatus)
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.POWER_TOGGLE(strCommand, tParams)
	g_powerStatus = ternary(g_powerStatus == 'ON', 'OFF', 'ON')
	ExecuteCommand(g_powerStatus)
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.RESTART(strCommand, tParams)
	dbg('CMDS.RESTART')
	g_powerStatus = 'OFF'
	ExecuteCommand('RESTART')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, g_powerStatus, {OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.SET_VOL_UP(strCommand, tParams)
	dbg('SET_VOL_UP')
	ExecuteCommand('SET_VOL_UP')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.SET_VOL_DOWN(strCommand, tParams)
	dbg('SET_VOL_DOWN')
	ExecuteCommand('SET_VOL_DOWN')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.DOLBY_DRC_ON(strCommand, tParams)
	dbg('DOLBY_DRC_ON')
	ExecuteCommand('DOLBY_DRC_ON')
end

function CMDS.DOLBY_DRC_OFF(strCommand, tParams)
	dbg('DOLBY_DRC_OFF')
	ExecuteCommand('DOLBY_DRC_OFF')
end

function CMDS.DOLBY_SCS_ON(strCommand, tParams)
	dbg('DOLBY_SCS_ON')
	ExecuteCommand('DOLBY_SCS_ON')
end

function CMDS.DOLBY_SCS_OFF(strCommand, tParams)
	dbg('DOLBY_SCS_OFF')
	ExecuteCommand('DOLBY_SCS_OFF')
end

function CMDS.DTS_DRC_ON(strCommand, tParams)
	dbg('DOLBY_DRC_ON')
	ExecuteCommand('DTS_DRC_ON')
end

function CMDS.DTS_DRC_OFF(strCommand, tParams)
	dbg('DOLBY_DRC_OFF')
	ExecuteCommand('DTS_DRC_OFF')
end

function CMDS.FRONT_PANEL_BRIGHTNESS_AUTO(strCommand, tParams)
	dbg('FRONT_PANEL_BRIGHTNESS_AUTO')
	ExecuteCommand('FRONT_PANEL_BRIGHTNESS_AUTO')
end

function CMDS.FRONT_PANEL_BRIGHTNESS_LOW(strCommand, tParams)
	dbg('FRONT_PANEL_BRIGHTNESS_LOW')
	ExecuteCommand('FRONT_PANEL_BRIGHTNESS_LOW')
end

function CMDS.FRONT_PANEL_BRIGHTNESS_MEDIUM(strCommand, tParams)
	dbg('FRONT_PANEL_BRIGHTNESS_MEDIUM')
	ExecuteCommand('FRONT_PANEL_BRIGHTNESS_MEDIUM')
end

function CMDS.FRONT_PANEL_BRIGHTNESS_HIGH(strCommand, tParams)
	dbg('FRONT_PANEL_BRIGHTNESS_HIGH')
	ExecuteCommand('FRONT_PANEL_BRIGHTNESS_HIGH')
end


function CMDS.SET_INPUT(strCommand, tParams)
	local inputID = tonumber(tParams.INPUT) % 1000
	if (inputID == g_currentInput) then
	   print('Duplicate Input Command Blocked')
	   return
     end
    
	g_currentInput = inputID
	ExecuteCommand('INPUT', DoubleDigit(inputID))
	ExecuteCommand('STSVOL')
	ExecuteCommand('STSMUT')
	C4:SendToProxy(5001, 'INPUT_OUTPUT_CHANGED', {INPUT = tParams.INPUT, OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'INPUT_OUTPUT_CHANGED', {INPUT = tParams.INPUT, OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'INPUT_OUTPUT_CHANGED', {INPUT = tParams.INPUT, OUTPUT = 2000}, 'NOTIFY')
	UpdateStatusProperties()
end

function CMDS.SET_SURROUND_MODE(strCommand, tParams)
	local mode = tSurroundCommands[tonumber(tParams['SURROUND_MODE'])]
	if (CMDS[mode]) then
		ExecuteCommand(mode)
	else
		ExecuteCommand('SOUND_MODE', mode)
	end
	C4:SendToProxy(5001, 'SURROUND_MODE_CHANGED', {SURROUND_MODE = mode, OUTPUT = 7001}, "NOTIFY")
	C4:SendToProxy(5001, 'SURROUND_MODE_CHANGED', {SURROUND_MODE = mode, OUTPUT = 4000}, "NOTIFY")
	C4:SendToProxy(5001, 'SURROUND_MODE_CHANGED', {SURROUND_MODE = mode, OUTPUT = 2000}, "NOTIFY")
	UpdateStatusProperties()
end

function CMDS.SET_VOLUME_LEVEL(strCommand, tParams)
	local level = tonumber(tParams.LEVEL)
	level = math.Clamp(level, 1, 99)
	g_inputVolume[g_currentInput] = level
	ExecuteCommand('VOLUME_DIRECT', DoubleDigit(level))
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
	UpdateStatusProperties()
end


function CMDS.SET_TREBLE_LEVEL(strCommand, tParams)
	local level = tonumber(tParams.LEVEL)
	level = math.Clamp(level, -10, 10)
	ExecuteCommand('TREBLE_LEVEL_DIRECT', ReturnSign(level) .. DoubleDigit(level))
	C4:SendToProxy(5001, 'TREBLE_LEVEL_CHANGED', {LEVEL = level, OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'TREBLE_LEVEL_CHANGED', {LEVEL = level, OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'TREBLE_LEVEL_CHANGED', {LEVEL = level, OUTPUT = 2000}, 'NOTIFY')
	g_trebelLevel = level
end

function CMDS.MUTE_ON(strCommand, tParams)
    dbg('function CMDS.MUTE_ON')
	g_inputMute[g_currentInput] = true
    ExecuteCommand('SET_MUTE_ON')
	ExecuteCommand('SET_MUTE_ON')
	C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
    C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
	UpdateStatusProperties()
end

function CMDS.MUTE_OFF(strCommand, tParams)
	dbg('function CMDS.MUTE_OFF')
    g_inputMute[g_currentInput] = false
    ExecuteCommand('SET_MUTE_OFF')
	ExecuteCommand('SET_MUTE_OFF')
	C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
    C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
	UpdateStatusProperties()
end


function CMDS.MUTE_TOGGLE(strCommand, tParams)
	dbg("function CMDS.MUTE_TOGGLE")
	if (g_inputMute[g_currentInput]) then
		CMDS.MUTE_OFF()
	else
		CMDS.MUTE_ON()
	end
	C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
    C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'MUTE_CHANGED', {MUTE = g_inputMute[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.SET_BASS_LEVEL(strCommand, tParams)
	local level = tonumber(tParams.LEVEL)
	level = math.Clamp(level, -10, 10)
	g_bassLevel = level
	ExecuteCommand('BASS_LEVEL_DIRECT', ReturnSign(level) .. DoubleDigit(level))
	C4:SendToProxy(5001, 'BASS_LEVEL_CHANGED', {LEVEL = level, OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'BASS_LEVEL_CHANGED', {LEVEL = level, OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'BASS_LEVEL_CHANGED', {LEVEL = level, OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.SET_LIPSYNC_DELAY(strCommand, tParams)
	local delay = math.Clamp(tonumber(tParams.DELAY), 0, 50)
	dbg('SET_LIPSYNC_DELAY: ' ..delay)
	ExecuteCommand('SET_LIPSYNC_DELAY', DoubleDigit(delay))
end

function CMDS.PULSE_VOL_UP(strCommand, tParams)
    g_inputVolume[g_currentInput] = math.Clamp(tonumber(g_inputVolume[g_currentInput]) + 1, 1, 99)
    ExecuteCommand('VOLUME_DIRECT', DoubleDigit(g_inputVolume[g_currentInput]))
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.PULSE_VOL_DOWN(strCommand, tParams)
    g_inputVolume[g_currentInput] = math.Clamp(tonumber(g_inputVolume[g_currentInput]) - 1, 1, 99)
    ExecuteCommand('VOLUME_DIRECT', DoubleDigit(g_inputVolume[g_currentInput]))
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.START_VOL_UP(strCommand, tParams)
    g_VolUpTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS')
    g_inputVolume[g_currentInput] = math.Clamp(tonumber(g_inputVolume[g_currentInput]) + 1, 1, 99)
    ExecuteCommand('VOLUME_DIRECT', DoubleDigit(g_inputVolume[g_currentInput]))
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.STOP_VOL_UP(strCommand, tParams)
    C4:KillTimer(g_VolUpTimer)
end

function CMDS.START_VOL_DOWN(strCommand, tParams)
    g_VolDownTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS')
    g_inputVolume[g_currentInput] = math.Clamp(tonumber(g_inputVolume[g_currentInput]) - 1, 1, 99)
    ExecuteCommand('VOLUME_DIRECT', DoubleDigit(g_inputVolume[g_currentInput]))
    C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
end

function CMDS.STOP_VOL_DOWN(strCommand, tParams)
    C4:KillTimer(g_VolDownTimer)
end

function CMDS.SET_DTS_DIALOG_CONTROL(strCommand, tParams)
    local value = tonumber(tParams.VALUE)
    value = math.Clamp(value, 0, 6)
    ExecuteCommand('DTS_DIALOG_CONTROL', value)
end

function CMDS.NUMBER_0(strCommand, tParams)
    NumberButton('0')
end

function CMDS.NUMBER_1(strCommand, tParams)
    NumberButton('1')
end

function CMDS.NUMBER_2(strCommand, tParams)
    NumberButton('2')
end

function CMDS.NUMBER_3(strCommand, tParams)
    NumberButton('3')
end

function CMDS.NUMBER_4(strCommand, tParams)
    NumberButton('4')
end

function CMDS.NUMBER_5(strCommand, tParams)
    NumberButton('5')
end

function CMDS.NUMBER_6(strCommand, tParams)
    NumberButton('6')
end

function CMDS.NUMBER_7(strCommand, tParams)
    NumberButton('7')
end

function CMDS.NUMBER_8(strCommand, tParams)
    NumberButton('8')
end

function CMDS.NUMBER_9(strCommand, tParams)
    NumberButton('9')
end

function CMDS.ENTER(strCommand, tParams)
    g_volumeString = ''
end

----------------------
-- HELPER FUNCTIONS --
----------------------

-- Ternary Operator
function ternary (cond , T, F)
    if cond then return T else return F end
end

-- Manages the Volume String for 10 Key Entry
function NumberButton(num)
    if (#g_volumeString == 2) then
	   g_volumeString = ''
    end
    g_volumeString = g_volumeString .. num
    local value = tonumber(g_volumeString)
    ExecuteCommand('VOLUME_DIRECT', DoubleDigit(value))
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 7001}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 4000}, 'NOTIFY')
	C4:SendToProxy(5001, 'VOLUME_LEVEL_CHANGED', {LEVEL = g_inputVolume[g_currentInput], OUTPUT = 2000}, 'NOTIFY')
end

-- Returns int of number clamp
function math.Clamp(val, lower, upper)
    if lower > upper then lower, upper = upper, lower end
    return math.max(lower, math.min(upper, val))
end

-- Returns number as string with two digit format (5 -> 05)
function DoubleDigit(val)
    local ret = tostring(val)
    if (#ret == 1) then
	   ret = "0"..ret
    end
    return ret
end

-- Returns string of sign
function ReturnSign(val)
    if (val >= 0) then
	   return "P"
    else
	   return "M"
    end
end

-- Returns decimal value of hex string
function fromHex(hexString)
   local stringLen = string.len(hexString)
   local decValue = 0
   local hexmap = {
    ['0'] = 0, ['1'] = 1,
    ['2'] = 2, ['3'] = 3,
    ['4'] = 4, ['5'] = 5,
    ['6'] = 6, ['7'] = 7,
    ['8'] = 8, ['9'] = 9,
    ['A'] = 10, ['B'] = 11,
    ['C'] = 12, ['D'] = 13,
    ['E'] = 14, ['F'] = 15}
    
   for k=0,stringLen-1 do
      lastdigit = string.sub(hexString, stringLen-k)
      decValue = decValue + (hexmap[lastdigit] * 16^k)
      hexString=string.sub(hexString,1,stringLen-1)
   end
   return decValue
end

-- Returns string hex code from string
function StringToHex(str)
    local output = ''
    local hexTable = {
	   [' '] = '20', ['#'] = '23',
	   ['0'] = '30', ['1'] = '31', ['2'] = '32',
	   ['3'] = '33', ['4'] = '34', ['5'] = '35',
	   ['6'] = '36', ['7'] = '37', ['8'] = '38',
	   ['9'] = '39', ['A'] = '41', ['B'] = '42',
	   ['C'] = '43', ['D'] = '44', ['E'] = '45',
	   ['F'] = '46', ['G'] = '47', ['H'] = '48',
	   ['I'] = '49', ['J'] = '4A', ['K'] = '4B',
	   ['L'] = '4C', ['M'] = '4D', ['N'] = '4E',
	   ['O'] = '4F', ['P'] = '50', ['Q'] = '51',
	   ['R'] = '52', ['S'] = '53', ['T'] = '54',
	   ['U'] = '55', ['V'] = '56', ['W'] = '57',
	   ['X'] = '58', ['Y'] = '59', ['Z'] = '5A' }
    for i=1, #str do
	   output = output .. hexTable[str:sub(i, i)] .. ' '
    end
    return output
end

function table.contains(tbl, item)
    return tbl[item] ~= nil
end

---------------------------
-- Async Port and Server --
---------------------------

function CreateNewTCPServer()
	local server = {
		clients = {},
		clientsCnt = 0,
		stop = function(self)
			local clients = self.clients
			self.clients = {}
			self.clientsCnt = 0

			if (self.socket ~= nil) then
				self.socket:Close()
				self.socket = nil
 
			for cli,info in pairs(clients) do
				print("Disconnecting " .. cli:GetRemoteAddress().ip .. ":" .. cli:GetRemoteAddress().port .. ": name: " .. tostring(info.name))
				cli:Write("Server is shutting down!\r\n"):Close(true)
				end
			end
		end,

		start = function(self, maxClients, bindAddr, port, done)
			local calledDone = false
			self.socket = C4:CreateTCPServer()
				:OnResolve(function(srv, endpoints)
					print("Server " .. tostring(srv) .. " resolved listening address")
					for i = 1, #endpoints do
						print("Available endpoints: [" .. i .. "] ip=" .. endpoints[i].ip .. ":" .. endpoints[i].port)
					end
				end)

				:OnListen(function(srv, endpoint)
					-- Handling this callback is optional.  It merely lets you know that the server is now actually listening.
					local addr = srv:GetLocalAddress()
					print("Server " .. tostring(srv) .. " chose endpoint " .. endpoint.ip .. ":" .. endpoint.port .. ", listening on " .. addr.ip .. ":" .. addr.port)
					if (not calledDone) then
						calledDone = true
						done(true, addr)
					end
					g_asyncPort = tostring(addr.port)
					C4:UpdateProperty('Async Port', g_asyncPort)
					ExecuteAsyncPairingCommand()
				end)

				:OnError(function(srv, code, msg, op)
					-- code is the system error code (as a number)
					-- msg is the error message as a string
					print("Server " .. tostring(srv) .. " Error " .. code .. " (" .. msg .. ")")
					if (not calledDone) then
						calledDone = true
						done(false, msg)
					end
				end)

				:OnAccept(function(srv, client)
					-- srv is the instance C4:CreateTCPServer() returned
					-- client is a C4LuaTcpClient instance of the new connection that was just accepted
					print("Connection on server " .. tostring(srv) .. " accepted, client: " .. tostring(client))
					local info = {}
					self.clients[client] = info
					self.clientsCnt = self.clientsCnt + 1
					client

				:OnRead(function(cli, data)
					print("Data Incomming from " .. cli:GetRemoteAddress().ip .. ":" .. cli:GetRemoteAddress().port .. ": " .. data)
					-- cli is the C4LuaTcpClient instance (same as client in the OnAccept handler) that the data was read on
					if (string.sub(data, -2) == "\r\n") then
						-- Need to check if the delimiter exists.  It may not if the client sent data without one and then disconnected!
						data = string.sub(data, 1, -3) -- Cut off \r\n
					end
					if (#data > 0) then
						ProcessMessage(data)
					end
					cli:ReadUntil("\r\n")
				end)

				:OnWrite(function(cli)
					-- cli is the C4LuaTcpClient instance (same as client in the OnAccept handler).  This callback is called when
					-- all data was sent.
					print("Server " .. tostring(srv) .. " Client " .. tostring(client) .. " Data was sent.")
				end)

				:OnDisconnect(function(cli, errCode, errMsg)
					-- cli is the C4LuaTcpClient instance (same as client in the OnAccept handler) that the data was read on
					-- errCode is the system error code (as a number).  On a graceful disconnect, this value is 0.
					-- errMsg is the error message as a string.
					if (errCode == 0) then
						print("Server " .. tostring(srv) .. " Client " .. tostring(client) .. " Disconnected gracefully.")
					else
						print("Server " .. tostring(srv) .. " Client " .. tostring(client) .. " Disconnected with error " .. errCode .. " (" .. errMsg .. ")")
					end
					self.clients[cli] = nil
					self.clientsCnt = self.clientsCnt - 1
				end)

				:OnError(function(cli, code, msg, op)
					-- cli is the C4LuaTcpClient instance (same as client in the OnAccept handler) that the data was read on
					-- code is the system error code (as a number)
					-- msg is the error message as a string
					-- op indicates what type of operation failed: "read", "write"
					print("Server " .. tostring(srv) .. " Client " .. tostring(client) .. " Error " .. code .. " (" .. msg .. ") on " .. op)
				end)

				:ReadUntil("\r\n")
			end)

			:Listen(bindAddr, port)

			if (self.socket ~= nil) then
					return self
			end
		end
 	}

	-- Start the server with a limit of 5 concurrent connections, listen on all interfaces on a randomly available port.  The server will shut down after 10 minutes.
	server:start(99, Properties['Director IP'], 0, function(success, info)
	-- Kind of silly that there is no C4 function to get the IP address of the director other than the obvious localhost at 127.0.0.1
	end)
end


--[[
Commands to send to pass along this port number and IP address of the controlling device.
"ASIP12" - 1st digit = 2
"ASIP24" - 2nd digit = 4
"ASIP36" - 3rd digit = 6
"ASIP48" - 4th digit = 8
"ASIP59" - 5th digit = 9
"ASIPCP" - Confirm Command : This pulls the IP address of the device that sent the command and loads the port #.
]]
function ExecuteAsyncPairingCommand()
	table.insert(tEmitCommands, 'ASIP1' .. string.sub(g_asyncPort, 1, 1))
	-- print('ASIP1' .. string.sub(g_asyncPort, 1, 1))
	table.insert(tEmitCommands, 'ASIP2' .. string.sub(g_asyncPort, 2, 2))
	-- print('ASIP2' .. string.sub(g_asyncPort, 2, 2))
	table.insert(tEmitCommands, 'ASIP3' .. string.sub(g_asyncPort, 3, 3))
	-- print('ASIP3' .. string.sub(g_asyncPort, 3, 3))
	table.insert(tEmitCommands, 'ASIP4' .. string.sub(g_asyncPort, 4, 4))
	-- print('ASIP4' .. string.sub(g_asyncPort, 4, 4))
	table.insert(tEmitCommands, 'ASIP5' .. string.sub(g_asyncPort, 5, 5))
	-- print('ASIP5' .. string.sub(g_asyncPort, 5, 5))
	table.insert(tEmitCommands, 'ASIPCP')
	idEmitCommandsTimer = C4:AddTimer(intraCommandDelayValue, 'MILLISECONDS')
end

function OnServerStatusChanged(nPort, strStatus)
    print("OnServerStatusChanged port: " .. nPort .. " status: " .. strStatus)
end

function OnServerConnectionStatusChanged(nHash, nPort, strStatus, strIP)
    print("OnServerConnectionStatusChanged hash: " .. nHash .. " port: " .. nPort .. " status: " .. strStatus .. " ip: " .. strIP)
end

function OnServerDataIn(nHandle, strData)
	print("OnServerDataIn" .. nHandle .. ": " .. strData)
	ProcessMessage(strData)
end

-- This Function Updates the Properties in Composer Pro
function UpdateStatusProperties()
	C4:UpdateProperty('Power Status', tostring(g_powerStatus))
	C4:UpdateProperty('Mute Status', tostring(g_inputMute[g_currentInput] and 'ON' or 'OFF'))
	C4:UpdateProperty('Input Status', tostring(g_currentInput))
	C4:UpdateProperty('Soundmode Status', tostring(g_soundMode))
	C4:UpdateProperty('Volume Status', tostring(g_inputVolume[g_currentInput]))
end

-----------------
-- INIT DRIVER --
-----------------

function OnDriverLateInit()
    C4:SendToSerial(1, 'VERBN2') -- Will Set 'Serial Available'
    C4:NetConnect(6003, Properties['Control Port'])
    C4:UpdateProperty('Director MAC Address', C4:GetUniqueMAC())
    C4:UpdateProperty('Director Version', C4:GetVersionInfo()['version'])
    
	CreateNewTCPServer() 
    ExecuteCommand('ON')
    print('Act4 Driver Succesfully Initialized')
end

print("Not bad, for a human!")